/**
 * Copyright 2024 SCOOP Software GmbH, cronn GmbH
 * SPDX-License-Identifier: AGPL-3.0-only
 */

"use client";

import {
  ApiInspectionFeature,
  ApiObjectType,
} from "@eshg/employee-portal-api/inspection";
import { optionsFromRecord } from "@eshg/lib-portal/components/formFields/SelectOptions";
import { addDays, formatISO } from "date-fns";
import { useMemo, useState } from "react";

import { procedureStatusNames } from "@/lib/baseModule/api/procedures/enums";
import { useGetPendingFacilities } from "@/lib/businessModules/inspection/api/queries/facility";
import { useIsNewFeatureEnabled } from "@/lib/businessModules/inspection/api/queries/feature";
import { useGetObjectTypes } from "@/lib/businessModules/inspection/api/queries/objectTypes";
import { ExportBannedFacilitiesButton } from "@/lib/businessModules/inspection/components/facility/pending/ExportBannedFacilitiesButton";
import { NewFacilityButton } from "@/lib/businessModules/inspection/components/facility/pending/NewFacilityButton";
import { PendingFacilitiesIncidentsSidebar } from "@/lib/businessModules/inspection/components/facility/pending/PendingFacilitiesIncidentsSidebar";
import { PotentialDuplicatesWarning } from "@/lib/businessModules/inspection/components/facility/pending/PotentialDuplicatesWarning";
import { useReviewFacilityDuplicateSidebar } from "@/lib/businessModules/inspection/components/facility/pending/ReviewFacilityDuplicateSidebar";
import { useReviewInspectionDuplicateSidebar } from "@/lib/businessModules/inspection/components/facility/pending/ReviewInspectionDuplicateSidebar";
import { ProcessImportButton } from "@/lib/businessModules/inspection/components/processImport/ProcessImportButton";
import {
  inspectionBannedFacilityFilterNames,
  inspectionDuplicateFilterNames,
  inspectionPendingFacilityKindNames,
  inspectionPhaseNames,
  inspectionTypeNames,
} from "@/lib/businessModules/inspection/shared/enums";
import { useIsOfflineFeatureEnabled } from "@/lib/businessModules/inspection/shared/offline/useIsOfflineFeatureEnabled";
import { PendingFacilitiesFilters } from "@/lib/businessModules/inspection/shared/types";
import { ButtonBar } from "@/lib/shared/components/buttons/ButtonBar";
import { FilterButton } from "@/lib/shared/components/buttons/FilterButton";
import { FilterSettings } from "@/lib/shared/components/filterSettings/FilterSettings";
import { FilterSettingsSheet } from "@/lib/shared/components/filterSettings/FilterSettingsSheet";
import { FilterDefinition } from "@/lib/shared/components/filterSettings/models/FilterDefinition";
import { FilterValue } from "@/lib/shared/components/filterSettings/models/FilterValue";
import {
  FilterSettingsStateProvider,
  useFilterSettings,
} from "@/lib/shared/components/filterSettings/useFilterSettings";
import { useSearchParamStateProvider } from "@/lib/shared/components/filterSettings/useSearchParamStateProvider";
import { Pagination } from "@/lib/shared/components/pagination/Pagination";
import { DataTable } from "@/lib/shared/components/table/DataTable";
import { TablePage } from "@/lib/shared/components/table/TablePage";
import { TableSheet } from "@/lib/shared/components/table/TableSheet";
import { TextInputFilter } from "@/lib/shared/components/tableFilters/TextInputFilter";
import { useTableControl } from "@/lib/shared/hooks/searchParams/useTableControl";

import {
  createPendingFacilitiesColumns,
  getPendingFacilityRowRoute,
} from "./columns";

type UserActivityState =
  | { type: "view-table" }
  | { type: "view-incidents"; inspectionId: string; facilityName: string };

const initialUserActivity: UserActivityState = { type: "view-table" };

function createFilterDefinitions(
  objectTypes: ApiObjectType[],
  isImportFeatureEnabled: boolean,
  isBannedFacilitiesExportFeatureEnabled: boolean,
): FilterDefinition[] {
  const objectTypeOptions = objectTypes.map((o) => ({
    label: o.name,
    value: o.id,
  }));

  const filterDefinitions: FilterDefinition[] = [
    {
      type: "EnumSingle",
      key: "kind",
      name: "Art",
      options: optionsFromRecord(inspectionPendingFacilityKindNames),
    },
    {
      type: "EnumSingle",
      key: "objectTypeId",
      name: "Objekttyp",
      options: objectTypeOptions,
    },
    {
      type: "Enum",
      key: "status",
      name: "Status",
      options: optionsFromRecord(procedureStatusNames),
    },
    {
      type: "Enum",
      key: "type",
      name: "Typ",
      options: optionsFromRecord(inspectionTypeNames),
    },
    {
      type: "Enum",
      key: "phase",
      name: "Phase",
      options: optionsFromRecord(inspectionPhaseNames),
    },
    {
      type: "Date",
      key: "isBefore",
      name: "Begehung vor",
    },
    {
      type: "Date",
      key: "isAfter",
      name: "Begehung nach",
    },
  ];

  if (isImportFeatureEnabled) {
    filterDefinitions.push({
      type: "EnumSingle",
      key: "hasDuplicates",
      name: "Duplikat",
      options: optionsFromRecord(inspectionDuplicateFilterNames),
    });
  }

  if (isBannedFacilitiesExportFeatureEnabled) {
    filterDefinitions.push({
      type: "EnumSingle",
      key: "banned",
      name: "Untersagte Einrichtung",
      options: optionsFromRecord(inspectionBannedFacilityFilterNames),
    });
  }

  return filterDefinitions;
}

export function PendingFacilitiesTable(
  props: Readonly<{ filter: PendingFacilitiesFilters }>,
) {
  const isOfflineEnabled = useIsOfflineFeatureEnabled();
  const isImportFeatureEnabled = useIsNewFeatureEnabled(
    ApiInspectionFeature.Import,
  );
  const isBannedFacilitiesExportFeatureEnabled = useIsNewFeatureEnabled(
    ApiInspectionFeature.BannedFacilitiesExport,
  );
  const { data: objectTypes } = useGetObjectTypes();

  const filterDefinitions = createFilterDefinitions(
    objectTypes,
    isImportFeatureEnabled,
    isBannedFacilitiesExportFeatureEnabled,
  );
  const paramStateProvider = useSearchParamStateProvider(
    filterDefinitions,
    true,
  );
  const { stateProvider, filter } = usePendingFacilitiesFilterState({
    stateProvider: paramStateProvider,
    defaults: [
      {
        type: "Enum",
        key: "phase",
        selectedValues: ["NEW", "PLANNING", "READY_FOR_EXECUTION"],
      },
      {
        type: "Date",
        key: "isBefore",
        selectedValue: formatISO(addDays(new Date(), 14), {
          representation: "date",
        }),
      },
    ],
    filter: props.filter,
  });

  function filterForDuplicates() {
    filterSettings.filterSettingsProps.activeFilterProps.deleteAllFilterValues();
    filterSettings.filterSettingsProps.onDraftValueChange("hasDuplicates", {
      type: "EnumSingle",
      key: "hasDuplicates",
      selectedValue: "true",
    });
    paramStateProvider.setActiveValues([
      {
        type: "EnumSingle",
        key: "hasDuplicates",
        selectedValue: "true",
      },
    ]);
  }

  const { data: procedures, isFetching } = useGetPendingFacilities(filter);

  const tableControl = useTableControl({ serverSideSorting: true });
  const columns = createPendingFacilitiesColumns(
    isOfflineEnabled,
    handleViewIncidentsClick,
    openReviewFacilityDuplicateSidebar,
    openReviewInspectionDuplicateSidebar,
    isImportFeatureEnabled,
  );

  const [userActivity, setUserActivity] =
    useState<UserActivityState>(initialUserActivity);

  const filterSettings = useFilterSettings({
    definitions: filterDefinitions,
    stateProvider,
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    onValuesSubmit: (_values) => {},
    showSearch: false,
  });

  const reviewFacilityDuplicateSidebar = useReviewFacilityDuplicateSidebar();
  const reviewInspectionDuplicateSidebar =
    useReviewInspectionDuplicateSidebar();

  function openReviewFacilityDuplicateSidebar(inspectionId: string) {
    reviewFacilityDuplicateSidebar.open({ inspectionId });
  }

  function openReviewInspectionDuplicateSidebar(inspectionId: string) {
    reviewInspectionDuplicateSidebar.open({ inspectionId });
  }

  function handleSidebarClosed() {
    setUserActivity(initialUserActivity);
  }

  function handleViewIncidentsClick(
    inspectionId: string,
    facilityName: string,
  ) {
    setUserActivity({
      type: "view-incidents",
      inspectionId: inspectionId,
      facilityName: facilityName,
    });
  }

  return (
    <>
      {isImportFeatureEnabled &&
        procedures.numberOfPossibleDuplicates !== 0 && (
          <PotentialDuplicatesWarning
            numberOfDuplicates={procedures.numberOfPossibleDuplicates}
            filterForDuplicates={filterForDuplicates}
          />
        )}
      <TablePage
        fullHeight
        controls={
          <ButtonBar
            left={
              <>
                <FilterButton {...filterSettings.filterButtonProps} />
                <TextInputFilter
                  searchParamName="name"
                  placeholder="Name"
                  tableControl={tableControl}
                />
                <TextInputFilter
                  searchParamName="postalCode"
                  placeholder="PLZ"
                  tableControl={tableControl}
                />
                <TextInputFilter
                  searchParamName="city"
                  placeholder="Stadt"
                  tableControl={tableControl}
                />
                <TextInputFilter
                  searchParamName="street"
                  placeholder="Straße"
                  tableControl={tableControl}
                />
              </>
            }
            right={
              <>
                <ExportBannedFacilitiesButton />
                <ProcessImportButton />
                <NewFacilityButton />
              </>
            }
          />
        }
        filterSettings={
          filterSettings.filterSettingsVisible && (
            <FilterSettingsSheet {...filterSettings.filterSettingsSheetProps}>
              <FilterSettings {...filterSettings.filterSettingsProps} />
            </FilterSettingsSheet>
          )
        }
      >
        <TableSheet
          loading={isFetching}
          footer={
            <Pagination
              totalCount={procedures.totalNumberOfElements}
              {...tableControl.paginationProps}
            />
          }
        >
          <DataTable
            data={procedures.elements}
            columns={columns}
            sorting={tableControl.tableSorting}
            rowNavigation={{
              route: getPendingFacilityRowRoute,
              focusColumnAccessorKey: "name",
            }}
            striped
          />
        </TableSheet>
      </TablePage>
      {userActivity.type === "view-incidents" && (
        <PendingFacilitiesIncidentsSidebar
          open={true}
          onClose={handleSidebarClosed}
          inspectionId={userActivity.inspectionId}
          facilityName={userActivity.facilityName}
        />
      )}
    </>
  );
}

function usePendingFacilitiesFilterState(options: {
  stateProvider: FilterSettingsStateProvider;
  filter: PendingFacilitiesFilters;
  defaults: FilterValue[];
}) {
  const { activeValues, setActiveValues, ...rest } = options.stateProvider;
  const [touched, setTouched] = useState(activeValues.length > 0);

  const stateProvider: FilterSettingsStateProvider = {
    activeValues: touched ? activeValues : options.defaults,
    setActiveValues: (values) => {
      setTouched(true);
      setActiveValues(values);
    },
    ...rest,
  };

  const filter: PendingFacilitiesFilters = useMemo(
    () => ({
      ...options.filter,
      ...(touched ? {} : activeValuesToFilters(options.defaults)),
    }),
    [options.filter, options.defaults, touched],
  );

  return {
    stateProvider,
    filter,
  };
}

function activeValuesToFilters(
  activeValues: FilterValue[],
): PendingFacilitiesFilters {
  const filters = new Map<string, unknown>();

  for (const value of activeValues) {
    switch (value.type) {
      case "Date":
      case "EnumSingle":
        filters.set(value.key, value.selectedValue);
        break;
      case "Enum":
        filters.set(value.key, value.selectedValues);
        break;
      case "Number":
        break;
    }
  }

  return Object.fromEntries(filters);
}
